from typing import List, Optional, Tuple, Union


class WebsocketMessage:
    kind: int
    data: Union[bytes, str]


SSLCtx = Tuple[bool, Optional[str], Optional[str], Optional[str], Optional[str], List[str], bool]



import click
import pytest

from granian.cli import Duration


@pytest.mark.parametrize(
    ('value', 'expected'),
    (
        ('10', 10),
        ('10s', 10),
        ('10m', 60 * 10),
        ('10m10s', 60 * 10 + 10),
        ('10h', 60 * 60 * 10),
        ('10d', 24 * 60 * 60 * 10),
        ('10d10h10m10s', 24 * 60 * 60 * 10 + 60 * 60 * 10 + 60 * 10 + 10),
    ),
)
def test_duration_convert(value: str, expected: int) -> None:
    duration_type = Duration()
    assert duration_type.convert(value, None, None) == expected


@pytest.mark.parametrize(
    ('value', 'error_message'),
    (
        ('10x', r"'10x' is not a valid duration"),
        ('10d10h10m10s10', r"'10d10h10m10s10' is not a valid duration"),
        ('10d10h10m10s10', r"'10d10h10m10s10' is not a valid duration"),
    ),
)
def test_duration_convert_invalid(value: str, error_message: str) -> None:
    duration_type = Duration()
    with pytest.raises(click.BadParameter, match=error_message):
        duration_type.convert(value, None, None)


@pytest.mark.parametrize(
    ('value', 'error_message'),
    (
        ('1000', r"'1000' is greater than the maximum allowed value of 100 seconds"),
        ('30m', r"'30m' is greater than the maximum allowed value of 100 seconds"),
        ('5', r"'5' is less than the minimum allowed value of 10 seconds"),
    ),
)
def test_duration_convert_out_of_range(value: str, error_message: str) -> None:
    duration_type = Duration(10, 100)
    with pytest.raises(click.BadParameter, match=error_message):
        duration_type.convert(value, None, None)



import asyncio
import os
import sys
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple


WrappableT = Callable[..., Any]
LoopBuilderT = Callable[..., asyncio.AbstractEventLoop]


class Registry:
    __slots__ = ['_data']

    def __init__(self):
        self._data: Dict[str, WrappableT] = {}

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def keys(self) -> Iterable[str]:
        return self._data.keys()

    def register(self, key: str) -> Callable[[WrappableT], WrappableT]:
        def wrap(builder: WrappableT) -> WrappableT:
            self._data[key] = builder
            return builder

        return wrap

    def get(self, key: str) -> Callable[..., Any]:
        try:
            return self._data[key]
        except KeyError:
            raise RuntimeError(f"'{key}' implementation not available.")


class BuilderRegistry(Registry):
    __slots__ = []

    def __init__(self):
        self._data: Dict[str, Tuple[LoopBuilderT, Dict[str, Any]]] = {}

    def register(self, key: str, packages: Optional[List[str]] = None) -> Callable[[LoopBuilderT], LoopBuilderT]:
        packages = packages or []

        def wrap(builder: LoopBuilderT) -> LoopBuilderT:
            loaded_packages, implemented = {}, True
            try:
                for package in packages:
                    __import__(package)
                    loaded_packages[package] = sys.modules[package]
            except ImportError:
                implemented = False
            if implemented:
                self._data[key] = (builder, loaded_packages)
            return builder

        return wrap

    def get(self, key: str) -> asyncio.AbstractEventLoop:
        try:
            builder, packages = self._data[key]
        except KeyError:
            raise RuntimeError(f"'{key}' implementation not available.")
        return builder(**packages)


loops = BuilderRegistry()


@loops.register('asyncio')
def build_asyncio_loop():
    loop = asyncio.new_event_loop() if os.name != 'nt' else asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)
    return loop


@loops.register('uvloop', packages=['uvloop'])
def build_uv_loop(uvloop):
    loop = uvloop.new_event_loop()
    asyncio.set_event_loop(loop)
    return loop


@loops.register('rloop', packages=['rloop'])
def build_rloop(rloop):
    loop = rloop.new_event_loop()
    asyncio.set_event_loop(loop)
    return loop


@loops.register('auto')
def build_auto_loop():
    if 'rloop' in loops:
        return loops.get('rloop')
    if 'uvloop' in loops:
        return loops.get('uvloop')
    return loops.get('asyncio')
